/*****************************************************************************
 * vout_pictures.c : picture management functions
 *****************************************************************************
 * Copyright (C) 2000 VideoLAN
 * $Id: vout_pictures.c,v 1.5 2001/12/19 03:50:22 sam Exp $
 *
 * Authors: Vincent Seguin <seguin@via.ecp.fr>
 *          Samuel Hocevar <sam@zoy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include "defs.h"

#include <errno.h>                                                 /* ENOMEM */
#include <stdlib.h>                                                /* free() */
#include <stdio.h>                                              /* sprintf() */
#include <string.h>                                            /* strerror() */

#include "common.h"
#include "intf_msg.h"
#include "threads.h"
#include "mtime.h"

#include "video.h"
#include "video_output.h"

/*****************************************************************************
 * vout_DisplayPicture: display a picture
 *****************************************************************************
 * Remove the reservation flag of a picture, which will cause it to be ready for
 * display. The picture won't be displayed until vout_DatePicture has been
 * called.
 *****************************************************************************/
void vout_DisplayPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );
    switch( p_picture->i_status )
    {
    case RESERVED_PICTURE:
        p_picture->i_status = RESERVED_DISP_PICTURE;
        break;
    case RESERVED_DATED_PICTURE:
        p_picture->i_status = READY_PICTURE;
        break;
#ifdef DEBUG
    default:
        intf_ErrMsg("error: picture %p has invalid status %d", p_picture, p_picture->i_status );
        break;
#endif
    }

#ifdef TRACE_VOUT
    intf_DbgMsg("picture %p", p_picture);
#endif
    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_DatePicture: date a picture
 *****************************************************************************
 * Remove the reservation flag of a picture, which will cause it to be ready
 * for display. The picture won't be displayed until vout_DisplayPicture has
 * been called.
 *****************************************************************************/
void vout_DatePicture( vout_thread_t *p_vout,
                       picture_t *p_picture, mtime_t date )
{
#ifdef TRACE_VOUT
    char        psz_date[ MSTRTIME_MAX_SIZE ];                       /* date */
#endif

    vlc_mutex_lock( &p_vout->picture_lock );
    p_picture->date = date;
    switch( p_picture->i_status )
    {
    case RESERVED_PICTURE:
        p_picture->i_status = RESERVED_DATED_PICTURE;
        break;
    case RESERVED_DISP_PICTURE:
        p_picture->i_status = READY_PICTURE;
        break;
#ifdef DEBUG
    default:
        intf_ErrMsg("error: picture %p has invalid status %d", p_picture, p_picture->i_status );
        break;
#endif
    }

#ifdef TRACE_VOUT
    intf_DbgMsg("picture %p, display date: %s", p_picture, mstrtime( psz_date, p_picture->date) );
#endif
    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_CreatePicture: allocate a picture in the video output heap.
 *****************************************************************************
 * This function creates a reserved image in the video output heap.
 * A null pointer is returned if the function fails. This method provides an
 * already allocated zone of memory in the picture data fields. It needs locking
 * since several pictures can be created by several producers threads.
 *****************************************************************************/
picture_t *vout_CreatePicture( vout_thread_t *p_vout,
                               boolean_t b_progressive,
                               boolean_t b_top_field_first,
                               boolean_t b_repeat_first_field )
{
    int         i_picture;                                  /* picture index */
    picture_t * p_picture;
    picture_t * p_free_picture = NULL;                 /* first free picture */

    /* Get lock */
    vlc_mutex_lock( &p_vout->picture_lock );

    /*
     * Look for an empty place. We start at 1 because the first
     * directbuffer is reserved for memcpy()ed pictures.
     */
    for( i_picture = 0;
         i_picture < I_RENDERPICTURES && p_free_picture == NULL;
         i_picture++ )
    {
        p_picture = PP_RENDERPICTURE[ i_picture ];

        /* If the picture we found is a memory buffer, and we might have
         * enough room later for a direct buffer, skip it. If no other
         * pictures are found, the video decoder will try again later. */
        if( p_vout->b_direct && ( p_vout->output.i_pictures > 3 )
             && ( p_picture->i_type != DIRECT_PICTURE ) )
        {
            break;
        }

        switch( p_picture->i_status )
        {
            case DESTROYED_PICTURE:
                /* Memory will not be reallocated, and function can end
                 * immediately - this is the best possible case, since no
                 * memory allocation needs to be done */
                p_picture->i_status = RESERVED_PICTURE;
                p_picture->i_refcount = 0;

                p_picture->b_progressive = b_progressive;
                p_picture->b_repeat_first_field = b_repeat_first_field;
                p_picture->b_top_field_first = b_top_field_first;

                p_vout->i_heap_size++;

                vlc_mutex_unlock( &p_vout->picture_lock );
                return( p_picture );

            case FREE_PICTURE:
                /* Picture is empty and ready for allocation */
                p_free_picture = p_picture;
                break;

            default:
                break;
        }
    }

    /*
     * Prepare picture
     */
    if( p_free_picture != NULL )
    {
        vout_AllocatePicture( p_free_picture,
                              p_vout->render.i_width, p_vout->render.i_height,
                              p_vout->render.i_chroma );

        if( p_free_picture->i_planes )
        {
            /* Copy picture information, set some default values */
            p_free_picture->i_status = RESERVED_PICTURE;
            p_free_picture->i_type = MEMORY_PICTURE;

            p_free_picture->i_refcount = 0;

            p_free_picture->b_progressive = b_progressive;
            p_free_picture->b_repeat_first_field = b_repeat_first_field;
            p_free_picture->b_top_field_first = b_top_field_first;

            p_free_picture->i_matrix_coefficients = 1;

            p_vout->i_heap_size++;
        }
        else
        {
            /* Memory allocation failed : set picture as empty */
            p_free_picture->i_status = FREE_PICTURE;
            p_free_picture = NULL;

            intf_ErrMsg( "vout error: picture allocation failed" );
        }

        vlc_mutex_unlock( &p_vout->picture_lock );

        /* Initialize mutex */
        vlc_mutex_init( &(p_free_picture->lock_deccount) );

        return( p_free_picture );
    }

    /* No free or destroyed picture could be found, but the decoder
     * will try again in a while. */
    vlc_mutex_unlock( &p_vout->picture_lock );

    return( NULL );
}

/*****************************************************************************
 * vout_DestroyPicture: remove a permanent or reserved picture from the heap
 *****************************************************************************
 * This function frees a previously reserved picture or a permanent
 * picture. It is meant to be used when the construction of a picture aborted.
 * Note that the picture will be destroyed even if it is linked !
 *****************************************************************************/
void vout_DestroyPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );

#ifdef DEBUG
    /* Check if picture status is valid */
    if( (p_picture->i_status != RESERVED_PICTURE) &&
        (p_picture->i_status != RESERVED_DATED_PICTURE) &&
        (p_picture->i_status != RESERVED_DISP_PICTURE) )
    {
        intf_ErrMsg( "error: picture %p has invalid status %d",
                     p_picture, p_picture->i_status );
    }
#endif

    p_picture->i_status = DESTROYED_PICTURE;
    p_vout->i_heap_size--;

    /* destroy the lock that had been initialized in CreatePicture */
    vlc_mutex_destroy( &(p_picture->lock_deccount) );

    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_LinkPicture: increment reference counter of a picture
 *****************************************************************************
 * This function increments the reference counter of a picture in the video
 * heap. It needs a lock since several producer threads can access the picture.
 *****************************************************************************/
void vout_LinkPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );
    p_picture->i_refcount++;

#ifdef TRACE_VOUT
    intf_DbgMsg( "picture %p refcount=%d", p_picture, p_picture->i_refcount );
#endif

    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_UnlinkPicture: decrement reference counter of a picture
 *****************************************************************************
 * This function decrement the reference counter of a picture in the video heap.
 *****************************************************************************/
void vout_UnlinkPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );
    p_picture->i_refcount--;

#ifdef TRACE_VOUT
    if( p_picture->i_refcount < 0 )
    {
        intf_DbgMsg( "error: refcount < 0" );
        p_picture->i_refcount = 0;
    }
#endif

    if( ( p_picture->i_refcount == 0 ) &&
        ( p_picture->i_status == DISPLAYED_PICTURE ) )
    {
        p_picture->i_status = DESTROYED_PICTURE;
        p_vout->i_heap_size--;
    }

#ifdef TRACE_VOUT
    intf_DbgMsg( "picture %p refcount=%d", p_picture, p_picture->i_refcount );
#endif

    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_RenderPicture: render a picture
 *****************************************************************************
 * This function chooses whether the current picture needs to be copied
 * before rendering, does the subpicture magic, and tells the video output
 * thread which direct buffer needs to be displayed.
 *****************************************************************************/
picture_t * vout_RenderPicture( vout_thread_t *p_vout, picture_t *p_picture,
                                                       subpicture_t *p_subpic )
{
    int i_index;

    if( p_picture == NULL )
    {
        /* XXX: subtitles */

        return NULL;
    }

    if( p_picture->i_type == DIRECT_PICTURE )
    {
        if( p_picture->i_refcount )
        {
            /* Picture is in a direct buffer and is still in use,
             * we need to copy it to another direct buffer before
             * displaying it if there are subtitles. */
            if( p_subpic != NULL )
            {
                    //printf("memcpy (refcount != 0)\n");
                /* We have subtitles. First copy the picture to
                 * the spare direct buffer, then render the
                 * subtitles. */
                for( i_index = 0 ; i_index < p_picture->i_planes ; i_index++ )
                {
                    p_main->fast_memcpy(
                        PP_OUTPUTPICTURE[0]->planes[ i_index ].p_data,
                        p_picture->planes[ i_index ].p_data,
                        p_picture->planes[ i_index ].i_bytes );
                }

                vout_RenderSubPictures( p_vout, PP_OUTPUTPICTURE[0], p_subpic );

                return PP_OUTPUTPICTURE[0];
            }

            /* No subtitles, picture is in a directbuffer so
             * we can display it directly even if it is still
             * in use. */
                    //printf("direct (refcount == 0)\n");
            return p_picture;
        }

        /* Picture is in a direct buffer but isn't used by the
         * decoder. We can safely render subtitles on it and
         * display it. */
                    //printf("direct (refcount == 0)\n");
        vout_RenderSubPictures( p_vout, p_picture, p_subpic );

        return p_picture;
    }

    /* Not a direct buffer. We either need to copy it to a direct buffer,
     * or render it if the chroma isn't the same. */
    if( p_vout->b_direct )
    {
        /* Picture is not in a direct buffer, but is exactly the
         * same size as the direct buffers. A memcpy() is enough,
         * then render the subtitles. */
                    //printf("memcpy (not a direct buffer)\n");
        for( i_index = 0; i_index < p_picture->i_planes; i_index++ )
        {
            p_main->fast_memcpy( PP_OUTPUTPICTURE[0]->planes[ i_index ].p_data,
                                 p_picture->planes[ i_index ].p_data,
                                 p_picture->planes[ i_index ].i_bytes );
        }

        vout_RenderSubPictures( p_vout, PP_OUTPUTPICTURE[0], p_subpic );

        return PP_OUTPUTPICTURE[0];
    }

    /* Picture is not in a direct buffer, and needs to be converted to
     * another size/chroma. Then the subtitles need to be rendered as
     * well. */

    /* This usually means software YUV, or hardware YUV with a
     * different chroma. */

                    //printf("render (not a direct buffer)\n");
    /* XXX: render to direct buffer */

    vout_RenderSubPictures( p_vout, p_picture, p_subpic );

    return &p_vout->p_picture[0];
}

/*****************************************************************************
 * vout_PlacePicture: calculate image window coordinates
 *****************************************************************************
 * This function will be accessed by plugins. It calculates the relative
 * position of the output window and the image window.
 *****************************************************************************/
void vout_PlacePicture( vout_thread_t *p_vout, int i_width, int i_height,
                        int *pi_x, int *pi_y, int *pi_width, int *pi_height )
{
    if( p_vout->b_scale )
    {
        *pi_width = i_width;
        *pi_height = i_height;
    }
    else
    {
        *pi_width = MIN( i_width, p_vout->render.i_width );
        *pi_height = MIN( i_height, p_vout->render.i_height );
    }

    if( VOUT_ASPECT_FACTOR * *pi_width / *pi_height < p_vout->render.i_aspect )
    {
        *pi_width = *pi_height * p_vout->render.i_aspect / VOUT_ASPECT_FACTOR;
    }
    else
    {
        *pi_height = *pi_width * VOUT_ASPECT_FACTOR / p_vout->render.i_aspect;
    }

    if( *pi_width > i_width )
    {
        *pi_width = i_width;
        *pi_height = VOUT_ASPECT_FACTOR * *pi_width / p_vout->render.i_aspect;
    }

    if( *pi_height > i_height )
    {
        *pi_height = i_height;
        *pi_width = *pi_height * p_vout->render.i_aspect / VOUT_ASPECT_FACTOR;
    }

    *pi_x = ( i_width - *pi_width ) / 2;
    *pi_y = ( i_height - *pi_height ) / 2;
}

/*****************************************************************************
 * vout_AllocatePicture: allocate a new picture in the heap.
 *****************************************************************************
 * This function allocates a fake direct buffer in memory, which can be
 * used exactly like a video buffer. The video output thread then manages
 * how it gets displayed.
 *****************************************************************************/
void vout_AllocatePicture( picture_t *p_picture,
                           int i_width, int i_height, int i_chroma )
{
    int i_data_size = 0;

    p_picture->i_size = i_width * i_height;

    /* Calculate coordinates */
    switch( i_chroma )
    {
        case YUV_420_PICTURE:        /* YUV 420: 1,1/4,1/4 samples per pixel */
            p_picture->i_chroma_size = p_picture->i_size / 4;
            p_picture->i_chroma_width = i_width / 2;
            break;

        case YUV_422_PICTURE:        /* YUV 422: 1,1/2,1/2 samples per pixel */
            p_picture->i_chroma_size = p_picture->i_size / 2;
            p_picture->i_chroma_width = i_width / 2;
            break;

        case YUV_444_PICTURE:            /* YUV 444: 1,1,1 samples per pixel */
            p_picture->i_chroma_size = p_picture->i_size;
            p_picture->i_chroma_width = i_width;
            break;

        default:
            intf_ErrMsg( "vout error: unknown chroma type %d", i_chroma );
            p_picture->i_planes = 0;
            return;
    }

    /* Allocate memory */
    switch( i_chroma )
    {
        case YUV_420_PICTURE:        /* YUV 420: 1,1/4,1/4 samples per pixel */
        case YUV_422_PICTURE:        /* YUV 422: 1,1/2,1/2 samples per pixel */
        case YUV_444_PICTURE:            /* YUV 444: 1,1,1 samples per pixel */

            i_data_size = p_picture->i_size + 2 * p_picture->i_chroma_size;

            /* The Y plane */
            p_picture->planes[ Y_PLANE ].i_bytes =
                 p_picture->i_size * sizeof(pixel_data_t);
            p_picture->planes[ Y_PLANE ].i_line_bytes =
                 i_width * sizeof(pixel_data_t);
            p_picture->planes[ Y_PLANE ].p_data =
                 memalign( 16, i_data_size * sizeof(pixel_data_t) * 4 );
            /* The U plane */
            p_picture->planes[ U_PLANE ].i_bytes =
                 p_picture->i_chroma_size * sizeof(pixel_data_t);
            p_picture->planes[ U_PLANE ].i_line_bytes =
                 p_picture->i_chroma_width * sizeof(pixel_data_t);
            p_picture->planes[ U_PLANE ].p_data =
                 p_picture->planes[ Y_PLANE ].p_data + p_picture->i_size;
            /* The V plane */
            p_picture->planes[ V_PLANE ].i_bytes =
                 p_picture->i_chroma_size * sizeof(pixel_data_t);
            p_picture->planes[ V_PLANE ].i_line_bytes =
                 p_picture->i_chroma_width * sizeof(pixel_data_t);
            p_picture->planes[ V_PLANE ].p_data =
                 p_picture->planes[ U_PLANE ].p_data + p_picture->i_chroma_size;

            p_picture->i_planes = 3;

            break;

        default:
            p_picture->i_planes = 0;

            break;
    }
}

